from pprint import pformat


def get_action_tags(istack, stack_tags):
    # cmd lines tags
    cmd_tags = {n.split("=")[0]: n.split("=")[1] for n in istack.cfg.tags}
    tags_cmd = {}
    tags_remove = {}

    # unchanged tags
    tags_default = {}

    # changed tags - same value as corresponding stack param
    tags_changed = {}

    # metadata tags - found inside template Metadata Section
    tags_metadata = {}

    final_tags = []

    for tag in stack_tags:
        key = tag["Key"]
        current_value = tag["Value"]

        # Skip LastUpdate and EnvApp1Version Tag
        if key in ["LastUpdate", "EnvApp1Version"]:
            continue

        # check if key exist as cfg param/attr too
        try:
            cfg_value = getattr(istack.cfg, key)
            in_cfg = True if cfg_value is not None else None
        except Exception:
            in_cfg = None

        # current value differ from cmd arg
        if in_cfg and current_value != cfg_value:
            value = cfg_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_changed[key] = "%s => %s" % (current_value, value)

        # remove tags using cmd --tags with tag value REMOVE
        elif key in cmd_tags and cmd_tags[key] == "REMOVE":
            tags_remove[key] = "REMOVE"
            continue
        # keep current tag value
        else:
            value = current_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_default[key] = value

        final_tags.append({"Key": key, "Value": value})

    final_tags_keys = [n["Key"] for n in final_tags]

    upsert_tags_cfg = [
        # Command line tags
        (cmd_tags, tags_cmd),
        # Metadata tags found inside template Metadata Section
        (istack.template.get("Metadata", {}).get("Tags", {}), tags_metadata),
    ]

    # Add or Update Tags
    for n in upsert_tags_cfg:
        for key, value in n[0].items():
            if key not in tags_remove:
                # Tag must not be removed
                tag = {"Key": key, "Value": value}
                if tag in final_tags:
                    # Tag already exist, skip it
                    pass
                elif key in final_tags_keys:
                    # Tag Key already exist but with different Value, update it
                    loc = final_tags_keys.index(key)
                    tags_changed[key] = "%s => %s" % (
                        final_tags[loc]["Value"],
                        value,
                    )
                    final_tags[loc] = tag
                    tags_default.pop(key, None)
                else:
                    # Tag Key do not exist, add/append it
                    final_tags.append(tag)
                    n[1][key] = value

    # Add LastUpdate Tag with current time
    # Currently disabled:
    # Some resource, like CloudFormation Distribution, take time to be updated.
    # Does it make sense to have a tag with LastUpdateTime even if resource properties are not changed at all ?
    # If a resource is created by CloudFormation i can simply look at Stack LastUpdateTime
    # to have the same information derived by tagging it (i know that with tagging is simpler to do this).
    # final_tags.append({"Key": "LastUpdate", "Value": str(datetime.now())})

    if tags_default:
        istack.mylog(
            "CURRENT - STACK TAGS\n%s\n" % pformat(tags_default, width=1000000)
        )
    if tags_changed:
        istack.mylog(
            "CHANGED - STACK TAGS\n%s\n" % pformat(tags_changed, width=1000000)
        )
    if tags_cmd:
        istack.mylog(
            "COMMAND LINE - STACK TAGS\n%s\n" % pformat(tags_cmd, width=1000000)
        )
    if tags_metadata:
        istack.mylog(
            "METADATA - STACK TAGS\n%s\n" % pformat(tags_metadata, width=1000000)
        )
    if tags_remove:
        istack.mylog("REMOVE - STACK TAGS\n%s\n" % pformat(tags_remove, width=1000000))

    return final_tags



# parser default cfg
stack = role = type = topics = stack_args = cmd_args = tags = []
parallel = region = jobs = pause = version = template = None
nowait = compact = dryrun = answer_yes = no_stacks = all_stacks = None
debug = False
max_retry_ecs_service_running_count = 0
timedelta = 300
dashboard = "OnChange"
statistic = "Average"
statisticresponse = "p95"
silent = True
vertical = False
profile = False
output = "text"
disable_rollback = False
dash_force = False
# changeset_original = False
print_mylog = True
#

OUT_WIDTH = 1000000

SLACK_CHANNEL = "_cf_deploy"

MAX_SINGLE_STACKS = 5

ACTION_WAITER_SLEEP_TIME = 3

STACK_BASE_DATA = [
    "StackName",
    "Description",
    "StackStatus",
    "CreationTime",
    "LastUpdatedTime",
]

RESOURCES_MAP = {
    "AWS::ApplicationAutoScaling::ScalableTarget": {
        "Name": "ClusterName",
        "PidEval": 'res_pid.split("/")[1]',
    },
    "AWS::ApplicationAutoScaling::ScalingPolicy": {
        "Name": "ScalingPolicyTrackingsECS",
        "PidEval": '"/".join(res_pid.split("/")[2:6])',
    },
    "AWS::AutoScaling::AutoScalingGroup": {
        "Name": "AutoScalingGroupName",
    },
    "AWS::AutoScaling::ScalingPolicy": {
        "Name": "ScalingPolicyTrackingsEC2",
        "PidEval": 'res_pid.split("/")[2]',
    },
    "AWS::CloudWatch::Alarm": {},
    "AWS::ECS::Cluster": {
        "Name": "ClusterName",
    },
    "AWS::ECS::Service": {
        "Name": "ServiceName",
        "PidEval": 'res_pid.split("/")[2]',
    },
    "AWS::ElasticLoadBalancing::LoadBalancer": {
        "Prefix": "LoadBalancerName",
    },
    "AWS::ElasticLoadBalancingV2::Listener": {
        "Prefix": "LoadBalancerListener",
        "PidEval": '"/".join(res_pid.split("/")[1:4])',
    },
    "AWS::ElasticLoadBalancingV2::ListenerRule": {
        "Prefix": "LoadBalancerListener",
        "PidEval": '"/".join(res_pid.split("/")[1:4])',
    },
    "AWS::ElasticLoadBalancingV2::LoadBalancer": {
        "Prefix": "LoadBalancer",
        "PidEval": '"/".join(res_pid.split("/")[1:4])',
    },
    "AWS::ElasticLoadBalancingV2::TargetGroup": {
        "Prefix": "TargetGroup",
        "PidEval": 'res_pid.split(":", 5)[5]',
    },
    "AWS::Route53::RecordSet": {},
    "AWS::ServiceDiscovery::Service": {},
    # LogicalResourceId
    "AlarmCPUHigh": {},
    "AlarmCPULow": {},
    "AutoScalingGroupSpot": {
        "Name": "AutoScalingGroupSpotName",
    },
    "ServiceSpot": {
        "Name": "ServiceSpotName",
    },
}

STACK_COMPLETE_STATUS = [
    "UPDATE_COMPLETE",
    "CREATE_COMPLETE",
    "ROLLBACK_COMPLETE",
    "UPDATE_ROLLBACK_COMPLETE",
    "UPDATE_ROLLBACK_FAILED",
    "DELETE_COMPLETE",
    "DELETE_FAILED",
]

CHANGESET_COMPLETE_STATUS = [
    "CREATE_COMPLETE",
    "UPDATE_ROLLBACK_FAILED",
    "FAILED",
]

SHOW_TABLE_FIELDS = [
    "EnvStackVersion",
    "EnvRole",
    "StackName",
    "StackType",
    "EnvApp1Version",
    "StackStatus",
    "LastUpdatedTime",
]

SHOW_RESOURCES_FIELDS = [
    "LogicalResourceId",
    "ResourceType",
    "ResourceStatus",
]

STACKSET_INSTANCES_SHOW_TABLE_FIELDS = [
    "Region",
    "Account",
    "StackId",
    "Status",
    "StatusReason",
    "StackInstanceStatus",
]

SSM_BASE_PATH = "/ibox"



import yaml
import logging


logging.basicConfig()
logging.getLogger("botocore").setLevel("CRITICAL")
logger = logging.getLogger("stacksops")
logger.setLevel(logging.INFO)


name = "iboxstacksops"
__version__ = "0.9.0"


class IboxError(Exception):
    pass


class IboxErrorECSService(Exception):
    pass


def yaml_exclamation_mark(dumper, data):
    if data.startswith(("!Ref", "!GetAtt", "!GetAZs")):
        tag = data.split(" ")[0]
        value = dumper.represent_scalar(tag, data.replace(f"{tag} ", ""))
    else:
        value = dumper.represent_scalar("tag:yaml.org,2002:str", data)

    return value


yaml.add_representer(str, yaml_exclamation_mark)



from . import (
    logger,
    cfg,
    template,
    parameters,
    resolve,
    actions,
    outputs,
    dashboard,
    ssm,
    route53,
)
from .aws import myboto3
from .tools import smodule_to_class
from .msg import msg_init


class ibox_stack(object):
    def __init__(self, name, base_data, region=None):
        # aws clients/resource
        self.boto3 = myboto3(self, region)
        self.cloudformation = self.boto3.resource("cloudformation")
        self.s3 = self.boto3.client("s3")
        self.client = self.boto3.client("cloudformation")

        # set property
        self.name = name
        self.bdata = base_data
        self.stack = None

        if isinstance(base_data, dict):
            for n, v in base_data.items():
                setattr(self, n, v)

        # self.cfg should contains parsed args
        # inside method processed by istack (in a parallel way)
        # i need to set attr to self.cfg and not to the common cfg
        self.cfg = smodule_to_class(cfg)

    def create(self):
        self.exports = self.cfg.exports
        self.template = template.get_template(self)
        self.c_parameters = {}
        parameters.process(self)
        resolve.process(self)
        result = actions.create(self)
        if result:
            return {self.name: self.stack.stack_status}

    def update(self):
        self.stack = self.cloudformation.Stack(self.name)
        self.exports = self.cfg.exports
        self.template = template.get_template(self)
        parameters.process(self)
        resolve.process(self)
        result = actions.update(self)

        if result:
            self.stack.reload()
            return self.stack.stack_status

    def delete(self):
        self.stack = self.cloudformation.Stack(self.name)
        actions.delete(self)

    def cancel_update(self):
        self.stack = self.cloudformation.Stack(self.name)
        result = actions.cancel_update(self)

        if result:
            self.stack.reload()
            return self.stack.stack_status

    def continue_update(self):
        self.stack = self.cloudformation.Stack(self.name)
        result = actions.continue_update(self)

        if result:
            self.stack.reload()
            return self.stack.stack_status

    def rollback(self):
        self.stack = self.cloudformation.Stack(self.name)
        result = actions.rollback(self)

        if result:
            self.stack.reload()
            return self.stack.stack_status

    def parameters(self, check=None):
        self.exports = self.cfg.exports
        self.template = template.get_template(self)
        parser = parameters.get_stack_parameter_parser(self)
        if check:
            parameters.add_stack_params_as_args(self, parser)
            return self.stack_parsed_args, self.parameters
        else:
            logger.info(f"{self.name} Parameters:")
            parser.print_help()

    def info(self):
        self.stack = self.cloudformation.Stack(self.name)
        self.template = template.get_template(self)
        outputs_out = outputs.show(self, "before")
        parameters_out = parameters.show_override(self)

        return {
            "OUTPUTS": outputs_out,
            "PARAMETERS NOT DEFAULT": parameters_out,
        }

    def show_resources(self):
        self.stack = self.cloudformation.Stack(self.name)
        actions.show_resources(self)

    def log(self):
        self.stack = self.cloudformation.Stack(self.name)
        actions.log(self)

    def resolve(self):
        self.exports = self.cfg.exports
        self.template = template.get_template(self)
        parameters.process(self, show=None)
        resolve.show(self)

    def ssm(self):
        self.ssm = self.boto3.client("ssm")
        ssm.put_parameters(self, self.bdata)

    def replicate(self, ssm_map, iregion):
        self.cfg.exports = iregion.cfg.exports
        self.cfg.stacks = iregion.cfg.stacks
        # pprint(ssm_map)
        for n, v in ssm_map.items():
            if n.startswith(f"{self.name}/"):
                parameter = n.split("/")[1]
                setattr(self.cfg, parameter, v)

        result = getattr(self, f"{self.cfg.command_replicate}")()

        return result

    def mylog(self, msg, chat=True):
        message = f"{self.name} # {msg}"
        try:
            print(message)
        except IOError:
            pass
        # logger.info(message)
        client = msg_init(self)
        if client and chat:
            try:
                client.chat_postMessage(
                    channel=f"#{cfg.slack_channel}",
                    text=message,
                    username=self.cfg.MSG_USER,
                    icon_emoji=":robot_face:",
                )
            except Exception as e:
                logger.warning(f"Error sending message to channel: {e}")

    def dash(self):
        dashboard.add_stack(self)

    def r53(self):
        self.route53 = self.boto3.client("route53")
        result = route53.create(self)
        return result

    def stackset_update(self):
        cfg.fields = cfg.STACKSET_INSTANCES_SHOW_TABLE_FIELDS
        self.exports = self.cfg.exports
        self.template = template.get_template(self, stackset=True)
        self.stack = True
        parameters.process(self)
        resolve.process(self)
        result = actions.stackset_update(self)
        if result:
            return {self.name: None}

    def stackset_info(self):
        self.exports = self.cfg.exports
        self.template = template.get_template(self, stackset=True)
        self.stack = True
        if not self.cfg.compact:
            self.cfg.OUT_WIDTH = 80
        parameters.show_override(self)

    def stackset_parameters(self):
        self.exports = self.cfg.exports
        self.template = template.get_template(self, stackset=True)
        parser = parameters.get_stack_parameter_parser(self)
        logger.info(f"{self.name} Parameters:")
        parser.print_help()

    def stackset_show(self):
        actions.stackset_show(self)

    def stackset_instances(self):
        cfg.fields = cfg.STACKSET_INSTANCES_SHOW_TABLE_FIELDS
        cfg.output = "text"
        actions.stackset_instances(self)


def exec_command(name, data, command, region=None, **kwargs):
    istack = ibox_stack(name, data, region)

    return getattr(istack, command)(**kwargs)



from pprint import pprint

from . import cfg, stacks, i_stack, i_region, table, ssm
from .tools import concurrent_exec, get_exports, show_confirm


def create():
    name = cfg.stack[0]
    cfg.exports = get_exports()
    stack = i_stack.ibox_stack(name, {})
    result = stack.create()
    if result:
        pprint(result)

    return result


def update():
    w_stacks = stacks.get()
    cfg.stacks = list(w_stacks.keys())
    cfg.exports = get_exports()
    if len(w_stacks) > 1 and (cfg.role or cfg.type) and not cfg.dryrun:
        print("You are going to UPDATE the following stacks:")
        print(cfg.stacks)
        if not show_confirm():
            return
    result = concurrent_exec("update", w_stacks, i_stack)
    if not cfg.dryrun:
        pprint(result)

    return result


def delete():
    w_stacks = stacks.get()
    cfg.stacks = list(w_stacks.keys())
    print("You are going to DELETE the following stacks:")
    print(cfg.stacks)
    if not show_confirm():
        return
    result = concurrent_exec("delete", w_stacks, i_stack)
    pprint(result)

    return result


def cancel_update():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec("cancel_update", w_stacks, i_stack)
    pprint(result)

    return result


def continue_update():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec("continue_update", w_stacks, i_stack)
    pprint(result)

    return result


def rollback():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec("rollback", w_stacks, i_stack)
    pprint(result)

    return result


def parameters():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    concurrent_exec("parameters", w_stacks, i_stack)


def info():
    if not cfg.compact:
        cfg.OUT_WIDTH = 80
    w_stacks = stacks.get()
    result = concurrent_exec("info", w_stacks, i_stack)

    return result


def show_resources():
    w_stacks = stacks.get()
    concurrent_exec("show_resources", w_stacks, i_stack)


def log():
    name = cfg.stack[0]
    stack = i_stack.ibox_stack(name, {})
    stack.log()


def resolve():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec("resolve", w_stacks, i_stack)

    return result


def dash():
    w_stacks = stacks.get()
    cfg.dash_name = "_" + "_".join(cfg.stack)
    cfg.jobs = 1
    result = concurrent_exec("dash", w_stacks, i_stack)

    return result


def show_table():
    w_stacks = stacks.get()
    s_table = table.get(list(w_stacks.values()))
    print(s_table)

    return s_table


def ssm_setup():
    w_stacks = stacks.get(exit_if_empty=False)
    result = concurrent_exec("ssm_setup", {k: w_stacks for k in cfg.regions}, i_region)
    pprint(result)


def ssm_put():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    stacks_params = concurrent_exec("parameters", w_stacks, i_stack, **{"check": True})
    regions = ssm.get_setupped_regions()
    w_regions = cfg.regions if cfg.regions else regions
    result = concurrent_exec(
        "ssm_put", {k: stacks_params for k in w_regions if k in regions}, i_region
    )
    pprint(result)


def ssm_show():
    w_stacks = stacks.get()
    regions = ssm.get_setupped_regions()
    result = concurrent_exec("ssm_get", {k: w_stacks for k in regions}, i_region)
    result = ssm.show(result)
    print(result)


def r53():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec("r53", w_stacks, i_stack)
    pprint(result)


def replicate():
    regions = ssm.get_setupped_regions() if not cfg.regions else cfg.regions

    if cfg.no_replicate_current:
        try:
            regions.remove(cfg.boto3.region_name)
        except Exception:
            pass

    result = concurrent_exec("replicate", {k: {} for k in regions}, i_region)
    pprint(result)

    return result


def stackset():
    name = cfg.stack[0]
    stackset = stacks.get(stackset=True)
    cfg.exports = get_exports()
    stack = i_stack.ibox_stack(name, stackset[name])
    cmd = getattr(stack, f"stackset_{cfg.command_stackset}")
    result = cmd()

    return result



from . import cfg, ssm, replica
from .aws import myboto3
from .tools import smodule_to_class


class ibox_region(object):
    def __init__(self, name, base_data):
        # aws clients/resource
        self.boto3 = myboto3(self, name)
        self.ssm = self.boto3.client("ssm")

        # set property
        self.name = name
        self.bdata = base_data

        if isinstance(base_data, dict):
            for n, v in base_data.items():
                setattr(self, n, v)

        # self.cfg should contains parsed args
        # inside method processed by istack (in a parallel way)
        # i need to set attr to self.cfg and not to the common cfg
        self.cfg = smodule_to_class(cfg)

    def ssm_setup(self):
        result = ssm.setup(self)
        return result

    def ssm_get(self):
        result = ssm.get_by_path(self, cfg.SSM_BASE_PATH)
        return result

    def ssm_put(self):
        result = ssm.put(self)
        return result

    def mylog(self, msg):
        message = f"{self.name} # {msg}"
        try:
            print(message)
        except IOError:
            pass

    def replicate(self):
        self.ssm_map = ssm.get_by_path(self, cfg.SSM_BASE_PATH)
        action = getattr(replica, cfg.command_replicate)
        result = action(self)
        return result


def exec_command(name, data, command, region=None, **kwargs):
    iregion = ibox_region(name, data)

    return getattr(iregion, command)(**kwargs)



def set_changed(istack):
    istack.after["resources"] = get(istack)

    before = istack.before["resources"]
    after = istack.after["resources"]

    changed = {}
    for r, v in before.items():
        if r in after and v != after[r]:
            changed[r] = after[r]

    istack.changed["resources"] = changed


def get(istack, rtypes=None):
    # todo manage dash or not
    resources = {}
    res_list = list(istack.cfg.RESOURCES_MAP.keys())

    paginator = istack.client.get_paginator("list_stack_resources")
    response_iterator = paginator.paginate(StackName=istack.name)

    for r in response_iterator:
        for res in r["StackResourceSummaries"]:
            res_lid = res["LogicalResourceId"]
            res_type = res["ResourceType"]
            res_pid = res.get("PhysicalResourceId")

            if res_type in res_list and (not rtypes or res_type in rtypes):
                # match on ResourceType
                conf = istack.cfg.RESOURCES_MAP[res_type]
                name = conf.get("Name")
                prefix = conf.get("Prefix")
                pid_eval = conf.get("PidEval")

                if prefix:
                    for n in ["External", "Internal"]:
                        if n in res_lid:
                            name = f"{prefix}{n}"
                if pid_eval:
                    try:
                        res_pid = eval(pid_eval)
                    except Exception:
                        pass

                if res_lid in res_list:
                    # match on LogicalResourceId too
                    conf = istack.cfg.RESOURCES_MAP[res_lid]
                    name = conf.get("Name", res_lid)

                if name:
                    res_lid = name

                resources[res_lid] = res_pid

    return resources



from pprint import pformat
from datetime import datetime

from . import cfg
from .tools import stack_resource_to_dict


def show_changed(istack):
    before = istack.before["outputs"]
    after = get(istack.stack)

    changed = {}
    for o, v in after.items():
        if o in before and v != before[o]:
            changed[o] = f"{before[o]} => {v}"
        elif o not in before:
            changed[o] = v

    istack.changed["outputs"] = changed
    istack.cfg.OUT_WIDTH = 80
    show(istack, "changed")


# show stack current outputs as dict
def show(istack, when):
    outputs = getattr(istack, when)["outputs"]

    out = pformat(outputs, width=istack.cfg.OUT_WIDTH)

    if istack.cfg.print_mylog:
        istack.mylog(f"{when.upper()} - STACK OUTPUTS\n{out}\n")

    return outputs


def get(stack):
    outputs = {}

    try:
        stack["StackName"]
    except Exception:
        stack = stack_resource_to_dict(stack)

    try:
        s_outputs = stack["Outputs"]
    except Exception:
        pass
    else:
        for output in s_outputs:
            key = output["OutputKey"]
            value = output.get("OutputValue", None)
            outputs[key] = value

    for d in cfg.STACK_BASE_DATA:
        value = stack.get(d, None)
        if isinstance(value, datetime):
            value = value.strftime("%Y-%m-%d %X %Z")
        outputs[d] = value

    return outputs



import os

from . import cfg

try:
    import slack
except ModuleNotFoundError:
    HAVE_MSG = False
else:
    HAVE_MSG = True


def msg_init(stack=None):
    obj = stack.cfg if stack else cfg
    try:
        return obj.MSG_CLIENT
    except Exception:
        MSG_AUTH = os.environ.get("IBOX_SLACK_TOKEN")
        MSG_USER = os.environ.get("IBOX_SLACK_USER")
        if HAVE_MSG and MSG_AUTH and MSG_USER and getattr(obj, "slack_channel", None):
            obj.MSG_USER = MSG_USER
            obj.MSG_CLIENT = slack.WebClient(token=MSG_AUTH)
            return obj.MSG_CLIENT



import time
from datetime import datetime
from pprint import pformat
from calendar import timegm

from . import IboxErrorECSService


# show old and new service tasks during an update
def _show_service_update(istack, event, timedelta):
    # avoid showing current service log if is requested a stack past event
    if timedelta != "0":
        return

    pri_task_def = deps_before = rolloutState = None
    deps_len = 0
    max_retry = istack.cfg.max_retry_ecs_service_running_count
    client = istack.boto3.client("ecs")

    # get the current Stack TaskDefinitions to be deployed
    stack_tasks_defs = [
        res.physical_resource_id
        for res in istack.stack.resource_summaries.all()
        if res.resource_type == "AWS::ECS::TaskDefinition"
    ]

    # get cluster and service from service arn
    try:
        cluster_name = event.physical_resource_id.split("/")[1]
        service_name = event.physical_resource_id.split("/")[2]
    except Exception:
        istack.mylog(
            f"Unable to retrieve Cluster and Service name from event Physical Resource ID: {event.physical_resource_id}\n"
            "skipping Service deployment logging."
        )
        return

    while (
        pri_task_def not in stack_tasks_defs
        or deps_len > 1
        or rolloutState == "IN_PROGRESS"
    ):
        deps = {
            "PRIMARY": {},
            "ACTIVE": {},
            "INACTIVE": {},
            "DRAINING": {},
        }

        service = client.describe_services(
            cluster=cluster_name,
            services=[service_name],
        )["services"][0]

        # dep_task_def = service["taskDefinition"]
        deployments = service["deployments"]
        deps_len = len(deployments)
        last_updatedAt = None

        for dep in deployments:
            status = dep["status"]
            dep_updatedAt = dep.get("updatedAt")
            if dep_updatedAt and (not last_updatedAt or dep_updatedAt > last_updatedAt):
                last_updatedAt = dep_updatedAt
            for p in [
                "taskDefinition",
                "desiredCount",
                "runningCount",
                "pendingCount",
                "failedTasks",
                "rolloutState",
            ]:
                deps[status][p] = dep.get(p)

        pri_task_def = deps["PRIMARY"]["taskDefinition"]
        failedTasks = deps["PRIMARY"]["failedTasks"]
        rolloutState = deps["PRIMARY"]["rolloutState"]

        if str(deps) != deps_before:
            if last_updatedAt:
                istack.mylog(last_updatedAt.strftime("%Y-%m-%d %X"))
            deps_before = str(deps)
            for d in ["PRIMARY", "ACTIVE", "DRAINING"]:
                if "taskDefinition" in deps[d]:
                    deps[d]["taskDefinition"] = deps[d]["taskDefinition"].split("/")[-1]
            istack.mylog("PRIMARY: %s" % pformat(deps["PRIMARY"], width=1000000))
            istack.mylog("ACTIVE: %s" % pformat(deps["ACTIVE"], width=1000000))
            istack.mylog("DRAINING: %s\n" % pformat(deps["DRAINING"], width=1000000))

            # is update stuck ?
            if (
                pri_task_def in stack_tasks_defs
                and max_retry > 0
                and failedTasks >= max_retry
            ):
                istack.last_event_timestamp = event.timestamp
                raise IboxErrorECSService(
                    "ECS Service did not stabilize "
                    f"[{failedTasks} >= {max_retry}] - "
                    "cancelling update [ROLLBACK]"
                )

        time.sleep(5)


# get timestamp from last event available
def get_last_timestamp(istack):
    last_event = list(istack.stack.events.all().limit(1))[0]

    return last_event.timestamp


# show all events after specific timestamp and return last event timestamp
def show(istack, timestamp, timedelta="0"):
    event_iterator = istack.stack.events.all()
    event_list = []
    for event in event_iterator:
        if event.timestamp > timestamp:
            event_list.insert(0, event)
        else:
            break
    for event in event_list:
        logtime = timegm(event.timestamp.timetuple())
        istack.mylog(
            event.logical_resource_id
            + " "
            + event.resource_status
            + " "
            + str(datetime.fromtimestamp(logtime))
            + " "
            + str(event.resource_status_reason)
        )
        # show service depoyment logging
        if (
            event.resource_type == "AWS::ECS::Service"
            and event.resource_status == "UPDATE_IN_PROGRESS"
            and event.resource_status_reason is None
            and istack.stack.stack_status not in istack.cfg.STACK_COMPLETE_STATUS
        ):
            _show_service_update(istack, event, timedelta)

    if len(event_list) > 0:
        return event_list.pop().timestamp

    return timestamp



import json
import botocore
import time
from datetime import timedelta

from . import (
    logger,
    resources,
    changeset,
    events,
    outputs,
    dashboard,
    table,
    IboxErrorECSService,
    IboxError,
)
from .tags import get_action_tags
from .tools import show_confirm


# build all args for action
def _get_action_args(istack):
    us_args = {}
    us_args["StackName"] = istack.name
    us_args["Parameters"] = istack.action_parameters
    us_args["Tags"] = istack.action_tags
    us_args["Capabilities"] = [
        "CAPABILITY_IAM",
        "CAPABILITY_NAMED_IAM",
        "CAPABILITY_AUTO_EXPAND",
    ]

    # sns topic
    us_args["NotificationARNs"] = istack.cfg.topics

    # Handle policy during update
    if hasattr(istack.cfg, "policy") and istack.cfg.policy:
        action = ['"Update:%s"' % a for a in istack.cfg.policy.split(",")]
        action = "[%s]" % ",".join(action)
        us_args["StackPolicyDuringUpdateBody"] = (
            '{"Statement" : [{"Effect" : "Allow",'
            '"Action" :%s,"Principal": "*","Resource" : "*"}]}' % action
        )

    if istack.template_from == "Current":
        us_args["UsePreviousTemplate"] = True
    if istack.template_from == "S3":
        us_args["TemplateURL"] = istack.cfg.template
    if istack.template_from == "File":
        us_args["TemplateBody"] = json.dumps(istack.template)

    if istack.cfg.disable_rollback:
        us_args["DisableRollback"] = True

    return us_args


# wait update until complete showing events status
def _update_waiter(istack, timestamp=None):
    last_timestamp = timestamp if timestamp else istack.last_event_timestamp

    # return without waiting
    if istack.cfg.nowait:
        istack.stack.reload()
        return

    while True:
        try:
            istack.stack.reload()
        except botocore.exceptions.ClientError as e:
            print(e)

        if istack.stack.stack_status in istack.cfg.STACK_COMPLETE_STATUS:
            events.show(istack, last_timestamp)
            break

        time.sleep(istack.cfg.ACTION_WAITER_SLEEP_TIME)

        try:
            last_timestamp = events.show(istack, last_timestamp)
        except IboxErrorECSService as e:
            # ECS Service did not stabilize, cancel update [ROLLBACK]
            logger.warning(e.args[0])
            cancel_update(istack)

    print("\n")


# wait until all stackset instances finished updating
def _stackset_update_waiter(istack):
    pending_instances = {}
    previous_instances = {}
    is_pending = True
    while is_pending:
        instances = stackset_instances(istack, False)
        for n in instances:
            stack_id = n["StackId"]
            if any(n["StackInstanceStatus"] == s for s in ["PENDING", "RUNNING"]):
                pending_instances[stack_id] = n
            else:
                try:
                    del pending_instances[stack_id]
                except Exception:
                    pass
        if pending_instances and previous_instances != instances:
            print(table.get(list(pending_instances.values())))

        previous_instances = instances

        if pending_instances:
            time.sleep(5)
        else:
            is_pending = False

    stackset_instances(istack)


def create(istack):
    stack_tags = [
        {"Key": "Env", "Value": istack.cfg.Env},
        {"Key": "EnvRole", "Value": istack.cfg.EnvRole},
        {"Key": "EnvStackVersion", "Value": istack.cfg.version},
        # Currently disabled. Same considerations of commit 360a1c3 (LastUpdate).
        # Furthermore, EnvAppVersion should be unrelated to Stack resources configuration.
        # So it make no sense to tag stack resource with it; should be sufficient
        # to look at CF Stack parameter value.
        # {"Key": "EnvApp1Version", "Value": istack.cfg.EnvApp1Version},
    ]

    # set tags
    istack.action_tags = get_action_tags(istack, stack_tags)

    if not show_confirm():
        return

    # get final args for update
    us_args = _get_action_args(istack)

    response = istack.client.create_stack(**us_args)
    istack.mylog(f"{json.dumps(response)}\n")
    time.sleep(1)

    istack.stack = istack.cloudformation.Stack(istack.name)
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    _update_waiter(istack)

    return True


def update(istack):
    # set tags
    istack.action_tags = get_action_tags(istack, istack.stack.tags)

    # get final args for update
    us_args = _get_action_args(istack)

    outputs.show(istack, "before")

    # -if using changeset ...
    if not istack.cfg.nochangeset and (
        len(istack.cfg.stacks) == 1 or istack.cfg.dryrun
    ):
        changeset_ok = changeset.process(istack, us_args)
        if not changeset_ok:
            return

    istack.before["resources"] = resources.get(istack)
    istack.last_event_timestamp = events.get_last_timestamp(istack)

    # do stack update
    try:
        response = istack.client.update_stack(**us_args)
    except botocore.exceptions.ClientError as err:
        raise IboxError(err)
    istack.mylog(f"{json.dumps(response)}\n")
    time.sleep(1)

    # -show update status until complete
    _update_waiter(istack)

    # show changed outputs
    outputs.show_changed(istack)

    # update dashboard
    dashboard.update(istack)

    return True


def delete(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.stack.delete()
    istack.mylog(f"{json.dumps(response)}\n")
    # -show update status until complete
    _update_waiter(istack)

    return True


def cancel_update(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.stack.cancel_update()
    istack.mylog(f"{json.dumps(response)}\n")
    # -show update status until complete
    _update_waiter(istack)

    return True


def continue_update(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.client.continue_update_rollback(
        StackName=istack.name, ResourcesToSkip=istack.cfg.resources_to_skip
    )
    istack.mylog(f"{json.dumps(response)}\n")
    # -show update status until complete
    _update_waiter(istack)

    return True


def rollback(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.client.rollback_stack(StackName=istack.name)
    istack.mylog(f"{json.dumps(response)}\n")
    # -show update status until complete
    _update_waiter(istack)

    return True


def log(istack):
    last_timestamp = events.get_last_timestamp(istack)
    time_delta = int(istack.cfg.timedelta)

    if time_delta == 0:
        time_event = last_timestamp - timedelta(seconds=1)
        _update_waiter(istack, time_event)
    else:
        if time_delta < 30:
            time_delta = time_delta * 86400
        time_event = last_timestamp - timedelta(seconds=time_delta)
        events.show(istack, time_event, time_delta)


def show_resources(istack):
    l_res = []
    paginator = istack.client.get_paginator("list_stack_resources")
    response_iterator = paginator.paginate(StackName=istack.name)

    for r in response_iterator:
        for res in r["StackResourceSummaries"]:
            if (
                not istack.cfg.resource_types
                or res["ResourceType"] in istack.cfg.resource_types
            ):
                l_res.append(res)
    s_table = table.get(l_res)
    istack.mylog(f"STACK RESOURCES:\n{s_table}")


def stackset_update(istack):
    # set tags
    istack.action_tags = get_action_tags(istack, istack.Tags)

    # get final args for update
    us_args = _get_action_args(istack)
    del us_args["StackName"]
    del us_args["NotificationARNs"]
    us_args["StackSetName"] = istack.name

    # do stack update
    # pprint(us_args)
    # exit(0)
    response = istack.client.update_stack_set(**us_args)
    istack.mylog(f"{json.dumps(response)}\n")
    time.sleep(2)

    _stackset_update_waiter(istack)

    return True


def stackset_instances(istack, show=True):
    response = istack.client.list_stack_instances(StackSetName=istack.name)

    for n, v in enumerate(response["Summaries"]):
        response["Summaries"][n]["StackId"] = response["Summaries"][n]["StackId"].split(
            "/"
        )[-2]
        response["Summaries"][n]["StackInstanceStatus"] = response["Summaries"][n][
            "StackInstanceStatus"
        ]["DetailedStatus"]

    if show:
        s_table = table.get(response["Summaries"])
        print(f"\nStackSetId: {istack.StackSetId}")
        print(s_table)

    return response["Summaries"]


def stackset_show(istack):
    s_table = table.get([istack.bdata])
    print(s_table)



import setuptools

setuptools.setup(
    packages=[
        "iboxstacksops",
    ],
    package_data={},
    install_requires=[
        "boto3",
        "prettytable",
        "PyYAML>=5,==5.*",
    ],
    extras_require={
        "extra": ["slackclient"],
    },
    python_requires=">=3.7",
    scripts=[
        "scripts/ibox_stacksops.py",
    ],
    include_package_data=True,
)



from prettytable import PrettyTable

from . import cfg


def get(data, fields=[]):
    table_data_names = {}

    for f in fields if fields else cfg.fields:
        if "=" in f:
            # custom ad hoc field can use = to split data name from column name
            loc = f.find("=")
            data_name = f[0:loc]
            field_name = f[loc + 1 :]
        else:
            data_name = field_name = f

        if any(data_name in n for n in data):
            table_data_names[field_name] = data_name

    fields = list(table_data_names)
    table_data = list(table_data_names.values())

    table = PrettyTable()
    table.padding_width = 1
    table.field_names = fields
    for n in data:
        parameter_not_empty = n.get("parameter_not_empty", [])
        table.add_row(
            [
                "Null"
                if (i not in n or not n[i])
                else f"{n[i]} *"
                if i in parameter_not_empty
                else n[i]
                for i in table_data
            ]
        )

    if fields:
        table.sortby = fields[0]
    table.reversesort = True
    table.align = "l"

    if cfg.output == "html":
        table.format = True

        return table.get_html_string(fields=fields)

    if cfg.output == "bare":
        table.header = False
        table.border = False
        table.left_padding_width = 0
        table.padding_width = 1

    return table.get_string(fields=fields)



#!/usr/bin/env python3
import sys

from iboxstacksops import IboxError, logger, cfg
from iboxstacksops.parser import set_cfg
from iboxstacksops.msg import msg_init


def main():
    set_cfg(sys.argv[1:])

    # Pre-Init msg client
    msg_init()

    try:
        cfg.func()
    except IboxError as e:
        logger.error(f"{e.args[0]}\n")
        return e


if __name__ == "__main__":
    result = main()

    if isinstance(result, IboxError):
        exit(1)



from pprint import pprint

from . import resources


def create(istack):
    def _get_sd_info(service_id):
        istack.servicediscovery = istack.boto3.client("servicediscovery")
        resp = istack.servicediscovery.get_service(Id=service_id)
        if resp["Service"]:
            service = resp["Service"]
            service_name = service["Name"]
            namespace_id = service["NamespaceId"]
            resp = istack.servicediscovery.get_namespace(Id=namespace_id)
            if resp["Namespace"]:
                namespace_name = resp["Namespace"]["Name"]

        return f"{service_name}.{namespace_name}"

    def _get_rec_info(record, rtype):
        r = {}
        param = record.split(".")
        r["stack"] = param[0]
        r["role"] = param[1]
        r["type"] = rtype
        if rtype == "external":
            r["region"] = param[2]
            r["domain"] = ".".join(param[3:6])
        if rtype == "internal":
            r["domain"] = ".".join(param[2:5])
        if rtype == "cf":
            del r["stack"]
            r["role"] = param[0]
            r["domain"] = ".".join(param[2:5])
        if rtype == "sd":
            r["region"] = param[2]
            r["domain"] = ".".join(param[3:])

        if istack.cfg.suffix and rtype != "cf":
            r["role"] = r["role"] + "-" + istack.cfg.suffix

        return r

    def _get_record_type(zoneid, name):
        resp = istack.route53.list_resource_record_sets(
            HostedZoneId=zoneid,
            StartRecordName=name,
            MaxItems="1",
        )

        if resp["ResourceRecordSets"]:
            return resp["ResourceRecordSets"][0]["Type"]
        else:
            return "A"

    def _get_record_change(name, zoneid, target, rtype):
        changes = {
            "Action": "UPSERT",
            "ResourceRecordSet": {
                "Name": name,
                "Type": rtype,
                "AliasTarget": {
                    "HostedZoneId": zoneid,
                    "DNSName": target,
                    "EvaluateTargetHealth": False,
                },
            },
        }

        return changes

    def _get_zoneid(domain):
        zones = istack.route53.list_hosted_zones_by_name(DNSName=domain)["HostedZones"]
        for z in zones:
            zoneid = z["Id"].split("/")[2]
            zone = istack.route53.get_hosted_zone(Id=zoneid)
            if zone["HostedZone"]["Name"] != domain + ".":
                continue
            try:
                zone_region = zone["VPCs"][0]["VPCRegion"]
            except Exception:
                return zoneid
            else:
                if zone_region == istack.boto3.region_name:
                    return zoneid

    res = resources.get(
        istack, rtypes=["AWS::Route53::RecordSet", "AWS::ServiceDiscovery::Service"]
    )
    pprint(res)
    out = {}
    for r, v in res.items():
        r_out = {}
        zoneid = None

        # process Route53 RecordSet
        if "External" in r:
            # External
            record = _get_rec_info(v, "external")
            record_region = "%s.%s.%s" % (
                record["role"],
                record["region"],
                record["domain"],
            )
            record_origin = "%s.origin.%s" % (record["role"], record["domain"])
            record_cf = "%s.%s" % (record["role"], record["domain"])
            map_record = {
                record_region: v,
            }

            if all("RecordSetCloudFront" not in n for n in res) and not istack.cfg.safe:
                map_record[record_cf] = record_region

            if not istack.cfg.noorigin and not istack.cfg.safe:
                map_record[record_origin] = record_region

        elif "Internal" in r:
            # Internal
            record = _get_rec_info(v, "internal")
            record_internal = record["role"] + "." + record["domain"]
            map_record = {
                record_internal: v,
            }

        elif "CloudFront" in r and not istack.cfg.safe:
            # CloudFront
            record = _get_rec_info(v, "cf")
            record_cf = record["role"] + "." + record["domain"]
            map_record = {
                record_cf: v,
            }

        elif "ServiceDiscoveryService" in r:
            # process ServiceDiscovery Service
            sd_record_name = _get_sd_info(v)
            record = _get_rec_info(sd_record_name, "sd")
            record_sd = record["role"] + "." + record["domain"]
            map_record = {
                record_sd: sd_record_name,
            }
            base_domain = ".".join(record["domain"].split(".")[1:])
            zoneid = _get_zoneid(base_domain)
        else:
            continue

        target_zoneid = _get_zoneid(record["domain"])

        if not zoneid:
            zoneid = target_zoneid

        for name, target in map_record.items():
            rtype = _get_record_type(target_zoneid, target)
            changes = _get_record_change(name, target_zoneid, target, rtype)
            print(name)
            pprint(changes)
            print("")

            if istack.cfg.dryrun:
                continue

            resp = istack.route53.change_resource_record_sets(
                HostedZoneId=zoneid, ChangeBatch={"Changes": [changes]}
            )
            pprint(resp["ChangeInfo"]["Status"])

            r_out[name] = target

        out[r] = r_out

    return out
